"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const anchor_1 = require("@coral-xyz/anchor");
const web3_js_1 = require("@solana/web3.js");
const fs_1 = __importDefault(require("fs"));
const mangoAccount_1 = require("../src/accounts/mangoAccount");
const client_1 = require("../src/client");
const constants_1 = require("../src/constants");
const utils_1 = require("../src/utils");
const CLUSTER_URL = process.env.CLUSTER_URL_OVERRIDE || process.env.MB_CLUSTER_URL;
const SOME_KEYPAIR = process.env.PAYER_KEYPAIR_OVERRIDE || process.env.MB_PAYER_KEYPAIR;
const CLUSTER = process.env.CLUSTER_OVERRIDE || 'mainnet-beta';
const GROUP_PK = '78b8f4cGCwmZ9ysPFMWLaLTkkaYnUjwMJYStWe5RTSSX';
async function main() {
    const options = anchor_1.AnchorProvider.defaultOptions();
    const connection = new web3_js_1.Connection(CLUSTER_URL, options);
    const someKeypair = web3_js_1.Keypair.fromSecretKey(Buffer.from(JSON.parse(fs_1.default.readFileSync(SOME_KEYPAIR, 'utf-8'))));
    const someWallet = new anchor_1.Wallet(someKeypair);
    const someProvider = new anchor_1.AnchorProvider(connection, someWallet, options);
    const client = client_1.MangoClient.connect(someProvider, CLUSTER, constants_1.MANGO_V4_ID[CLUSTER], {
        idsSource: 'api',
    });
    const group = await client.getGroup(new web3_js_1.PublicKey(GROUP_PK));
    const mangoAccountsWithHealth = (await client.getAllMangoAccounts(group, true))
        .map((a) => {
        return {
            account: a,
            healthRatio: a.getHealthRatioUi(group, mangoAccount_1.HealthType.maint),
            equity: (0, utils_1.toUiDecimalsForQuote)(a.getEquity(group)),
        };
    })
        .filter((a) => a.equity > 1000)
        .filter((a) => a.healthRatio < 50)
        .sort((a, b) => a.healthRatio - b.healthRatio);
    console.log(`${'Owner'.padStart(45)}, ${'Account'.padStart(45)}, ${'Health Ratio'.padStart(10)}, ${'Equity'.padStart(10)}`);
    for (const obj of mangoAccountsWithHealth) {
        console.log(`${obj.account.owner.toBase58().padStart(45)} ${obj.account.publicKey
            .toBase58()
            .padStart(45)}: ${obj.healthRatio
            .toFixed(2)
            .padStart(8)} %, ${obj.equity.toLocaleString().padStart(10)} $`);
    }
    process.exit();
}
try {
    main();
}
catch (error) {
    console.log(error);
}
