"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const anchor_1 = require("@coral-xyz/anchor");
const web3_js_1 = require("@solana/web3.js");
const console_1 = require("console");
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const perp_1 = require("../../src/accounts/perp");
const serum3_1 = require("../../src/accounts/serum3");
const client_1 = require("../../src/client");
const constants_1 = require("../../src/constants");
const utils_1 = require("../../src/utils");
const rpc_1 = require("../../src/utils/rpc");
const defaultParams = __importStar(require("./params/default.json"));
const sequence_enforcer_util_1 = require("./sequence-enforcer-util");
const MangoFillOrKill = __importStar(require("./mango_fill_or_kill"));
const tuple = (...args) => args;
console.log(defaultParams);
// Future
// * use async nodejs logging
// * merge gMa calls
// * take out spammers
// * batch ixs across various markets
// * only refresh part of the group which market maker is interested in
// Env vars
const CLUSTER = process.env.CLUSTER_OVERRIDE || 'mainnet-beta';
const CLUSTER_URL = process.env.CLUSTER_URL_OVERRIDE || process.env.MB_CLUSTER_URL;
const USER_KEYPAIR = process.env.USER_KEYPAIR_OVERRIDE || process.env.MB_PAYER_KEYPAIR;
const MANGO_ACCOUNT_PK = process.env.MANGO_ACCOUNT_PK || '';
// Load configuration
const paramsFileName = process.env.PARAMS || 'default.json';
const params = JSON.parse(fs_1.default.readFileSync(path_1.default.resolve(__dirname, `./params/${paramsFileName}`), 'utf-8'));
const control = { isRunning: true, interval: params.interval };
function getPerpMarketAssetsToTradeOn(group) {
    const allMangoGroupPerpMarketAssets = Array.from(group.perpMarketsMapByName.keys()).map((marketName) => marketName.replace('-PERP', ''));
    const allMangoGroupOpenBookMarketAssets = Array.from(group.serum3MarketsMapByMarketIndex.values()).map((market) => market.name.replace('/USDC', ''));
    return Object.keys(params.assets).filter((asset) => allMangoGroupPerpMarketAssets.includes(asset) &&
        allMangoGroupOpenBookMarketAssets.includes(asset));
}
// Refresh group, mango account and perp markets
async function refreshState(client, group, mangoAccount, marketContexts) {
    const ts = Date.now() / 1000;
    const result = await Promise.all([
        group.reloadAll(client),
        mangoAccount.reload(client),
    ]);
    Array.from(marketContexts.values()).map(async (mc, i) => {
        const perpMarket = mc.perpMarket;
        mc.perpMarket = group.getPerpMarketByMarketIndex(perpMarket.perpMarketIndex);
        mc.perpBids = await perpMarket.loadBids(client, true);
        mc.perpAsks = await perpMarket.loadAsks(client, true);
        const spotMarket = mc.spotMarket;
        mc.spotMarket = group.getSerum3MarketByMarketIndex(spotMarket.marketIndex);
        mc.spotBids = await spotMarket.loadBids(client, group);
        mc.spotAsks = await spotMarket.loadAsks(client, group);
        mc.lastBookUpdate = ts;
    });
    return {
        mangoAccount,
        lastMangoAccountUpdate: ts,
        marketContexts,
    };
}
// Initialiaze sequence enforcer accounts
async function initSequenceEnforcerAccounts(client, marketContexts) {
    const seqAccIxs = marketContexts.map((mc) => (0, sequence_enforcer_util_1.makeInitSequenceEnforcerAccountIx)(mc.sequenceAccount, client.program.provider.wallet.publicKey, mc.sequenceAccountBump, mc.perpMarket.name, CLUSTER));
    // eslint-disable-next-line
    while (true) {
        try {
            const sig = await (0, rpc_1.sendTransaction)(client.program.provider, seqAccIxs, []);
            console.log(`Sequence enforcer accounts created, sig https://explorer.solana.com/tx/${sig}?cluster=${CLUSTER == 'devnet' ? 'devnet' : ''}`);
        }
        catch (e) {
            console.log('Failed to initialize sequence enforcer accounts!');
            console.log(e);
            continue;
        }
        break;
    }
}
async function cancelAllOrdersForAMarket(client, group, mangoAccount, perpMarket) {
    for (const i of Array(100).keys()) {
        await (0, rpc_1.sendTransaction)(client.program.provider, [
            await client.perpCancelAllOrdersIx(group, mangoAccount, perpMarket.perpMarketIndex, 10),
        ], []);
        await mangoAccount.reload(client);
        if ((await mangoAccount.loadPerpOpenOrdersForMarket(client, group, perpMarket.perpMarketIndex)).length === 0) {
            break;
        }
    }
}
// Main driver for the market maker
async function fullMarketMaker() {
    // Load client
    const options = anchor_1.AnchorProvider.defaultOptions();
    const connection = new web3_js_1.Connection(CLUSTER_URL, options);
    const user = web3_js_1.Keypair.fromSecretKey(Buffer.from(JSON.parse(process.env.KEYPAIR || fs_1.default.readFileSync(USER_KEYPAIR, 'utf-8'))));
    const userWallet = new anchor_1.Wallet(user);
    const userProvider = new anchor_1.AnchorProvider(connection, userWallet, options);
    const client = await client_1.MangoClient.connect(userProvider, CLUSTER, constants_1.MANGO_V4_ID[CLUSTER], {
        idsSource: 'get-program-accounts',
    });
    const fillOrKill = new anchor_1.Program(MangoFillOrKill.IDL, '8CDZXDdwjrNJViXQWybrN27RzTc2aWSbFyYhQGTAxpAZ', userProvider);
    // Load mango account
    let mangoAccount = await client.getMangoAccount(new web3_js_1.PublicKey(MANGO_ACCOUNT_PK));
    console.log(`MangoAccount ${mangoAccount.publicKey} for user ${user.publicKey} ${mangoAccount.isDelegate(client) ? 'via delegate ' + user.publicKey : ''}`);
    await mangoAccount.reload(client);
    // Load group
    const group = await client.getGroup(mangoAccount.group);
    await group.reloadAll(client);
    // Build and maintain an aggregate context object per market
    const marketContexts = new Map();
    for (const perpMarketAsset of getPerpMarketAssetsToTradeOn(group)) {
        const perpMarket = group.getPerpMarketByName(perpMarketAsset + '-PERP');
        const spotMarket = group.getSerum3MarketByName(perpMarketAsset + '/USDC');
        const [sequenceAccount, sequenceAccountBump] = await web3_js_1.PublicKey.findProgramAddress([
            Buffer.from(perpMarket.name, 'utf-8'),
            client.program.provider.wallet.publicKey.toBytes(),
        ], sequence_enforcer_util_1.seqEnforcerProgramIds[CLUSTER]);
        marketContexts.set(perpMarket.perpMarketIndex, {
            params: params.assets[perpMarketAsset].perp,
            perpMarket: perpMarket,
            perpBids: await perpMarket.loadBids(client),
            perpAsks: await perpMarket.loadAsks(client),
            spotMarket: spotMarket,
            spotBids: await spotMarket.loadBids(client, group),
            spotAsks: await spotMarket.loadAsks(client, group),
            lastBookUpdate: 0,
            sequenceAccount,
            sequenceAccountBump,
        });
    }
    // Init sequence enforcer accounts
    await initSequenceEnforcerAccounts(client, Array.from(marketContexts.values()));
    // Load state first time
    console.log(`Loading state first time`);
    let state = await refreshState(client, group, mangoAccount, marketContexts);
    // Add handler for e.g. CTRL+C
    process.on('SIGINT', function () {
        console.log('Caught keyboard interrupt. Exiting...');
        control.isRunning = false;
    });
    // Loop indefinitely
    while (control.isRunning) {
        try {
            console.log(`\nRefreshing state`);
            refreshState(client, group, mangoAccount, marketContexts).then((result) => (state = result));
            mangoAccount = state.mangoAccount;
            for (const mc of marketContexts.values()) {
                const [summary, ixs] = await makeArbInstructions(client, fillOrKill, group, mangoAccount, mc);
                if (ixs.length === 0) {
                    continue;
                }
                const sig = await (0, rpc_1.sendTransaction)(client.program.provider, ixs, group.addressLookupTablesList);
                console.log(summary);
                console.log(`Executed trade, sig https://explorer.solana.com/tx/${sig}?cluster=${CLUSTER == 'devnet' ? 'devnet' : ''}`);
                // Abort after one!
                control.isRunning = false;
            }
        }
        catch (e) {
            console.log(e);
        }
        finally {
            console.log(`${new Date().toUTCString()} sleeping for ${control.interval / 1000}s`);
            await new Promise((r) => setTimeout(r, control.interval));
        }
    }
}
async function makeArbInstructions(client, fillOrKill, group, mangoAccount, mc) {
    const perpMarketIndex = mc.perpMarket.perpMarketIndex;
    const perpMarket = mc.perpMarket;
    const spotMarket = mc.spotMarket;
    const perpPos = mangoAccount.getPerpPositionUi(group, perpMarketIndex, true);
    const baseBank = group.banksMapByTokenIndex.get(spotMarket.baseTokenIndex)[0];
    const oo = mangoAccount.serum3OosMapByMarketIndex.get(spotMarket.marketIndex);
    const orderId = Date.now();
    const instructions = [
        (0, sequence_enforcer_util_1.makeCheckAndSetSequenceNumberIx)(mc.sequenceAccount, client.program.provider.wallet.publicKey, orderId, CLUSTER),
    ];
    instructions.push(await client.healthRegionBeginIx(group, mangoAccount, [], [perpMarket]));
    const fees = 0.0008; // 2 bps serum + 6 bps perp
    const threshold = 2 * fees;
    const l2Depth = 3;
    const maxAmount = 0.01;
    const maxPerpPos = 1.0;
    for (const [action, bids, asks] of [
        tuple('perp-short', mc.perpBids.getL2(l2Depth), mc.spotAsks.getL2(l2Depth)),
        tuple('perp-long', mc.spotBids.getL2(l2Depth), mc.perpAsks.getL2(l2Depth)),
    ]) {
        if (bids.length == 0 || asks.length == 0) {
            continue;
        }
        const topBid = bids[0];
        const topAsk = asks[0];
        // TODO: Could be much more elaborate on matching, not only looking at the first order
        if (topBid[0] < topAsk[0] * (1 + threshold)) {
            continue;
        }
        let amount = Math.min(topBid[1], topAsk[1], maxAmount);
        let summary = '';
        if (action == 'perp-short') {
            amount = Math.min(maxPerpPos + perpPos, amount);
            if (amount < perpMarket.minOrderSize) {
                continue;
            }
            const spotIxs = await client.serum3PlaceOrderIx(group, mangoAccount, spotMarket.serumMarketExternal, serum3_1.Serum3Side.bid, topAsk[0], amount, serum3_1.Serum3SelfTradeBehavior.decrementTake, serum3_1.Serum3OrderType.immediateOrCancel, orderId, 10);
            const perpIx = await client.perpPlaceOrderIx(group, mangoAccount, perpMarketIndex, perp_1.PerpOrderSide.ask, topBid[0], amount, undefined, orderId, perp_1.PerpOrderType.immediateOrCancel, false, undefined, 10);
            instructions.push(...spotIxs, perpIx);
            const checkPerpIx = await fillOrKill.methods
                .checkMangoV4PerpBase(perpMarketIndex, perpMarket.uiBaseToLots(perpPos - amount))
                .accounts({
                account: mangoAccount.publicKey,
            })
                .instruction();
            const nativeAmount = (0, utils_1.toNative)(amount, group.getMintDecimalsByTokenIndex(spotMarket.baseTokenIndex));
            const checkSpotIx = await fillOrKill.methods
                .checkOpenbookOpenOrdersBase(oo.baseTokenTotal.add(nativeAmount))
                .accounts({
                openOrders: oo.address,
            })
                .instruction();
            instructions.push(checkPerpIx, checkSpotIx);
            summary = `ACTION perp-short: ${amount},${topAsk[0]},${-amount},${topBid[0]}`;
        }
        else if (action == 'perp-long') {
            amount = Math.min(maxPerpPos - perpPos, amount);
            if (amount < perpMarket.minOrderSize) {
                continue;
            }
            const spotIxs = await client.serum3PlaceOrderIx(group, mangoAccount, spotMarket.serumMarketExternal, serum3_1.Serum3Side.ask, topBid[0], amount, serum3_1.Serum3SelfTradeBehavior.decrementTake, serum3_1.Serum3OrderType.immediateOrCancel, orderId, 10);
            const perpIx = await client.perpPlaceOrderIx(group, mangoAccount, perpMarketIndex, perp_1.PerpOrderSide.bid, topAsk[0], amount, undefined, orderId, perp_1.PerpOrderType.immediateOrCancel, false, undefined, 10);
            instructions.push(...spotIxs, perpIx);
            // TODO: check ix!
            summary = `ACTION perp-long: ${-amount},${topBid[0]},${amount},${topAsk[0]}`;
        }
        else {
            (0, console_1.assert)(false);
        }
        instructions.push(await client.healthRegionEndIx(group, mangoAccount, [], [perpMarket]));
        return [summary, instructions];
    }
    return ['', []];
}
function startMarketMaker() {
    try {
        if (control.isRunning) {
            fullMarketMaker()
                .catch((error) => console.log(error))
                .finally(startMarketMaker);
        }
    }
    catch (error) {
        console.log(error);
    }
}
startMarketMaker();
